# -*- coding: utf-8 -*-
# Part of Odoo. See LICENSE file for full copyright and licensing details.

from collections import defaultdict

from odoo import api, fields, models


class SaleOrder(models.Model):
    _inherit = 'sale.order'

    referrer_id = fields.Many2one('res.partner', 'Referrer', domain=[('grade_id', '!=', False)])
    commission_plan_id = fields.Many2one(
        'commission.plan',
        'Commission Plan',
        compute='_compute_commission_plan',
        inverse='_set_commission_plan',
        store=True,
        tracking=True,
        help="Takes precedence over the Referrer's commission plan.")
    commission = fields.Monetary(string='Referrer Commission', compute='_compute_commission')

    @api.depends('referrer_id', 'commission_plan_id', 'pricelist_id', 'order_line.price_total')
    def _compute_commission(self):
        for so in self:
            if not so.referrer_id or not so.commission_plan_id:
                so.commission = 0
            else:
                comm_by_rule = defaultdict(float)

                for line in so.order_line:
                    template = line.subscription_id.template_id
                    template_id = template.id if template else None
                    rule = so.commission_plan_id._match_rules(line.product_id, template_id, so.pricelist_id.id)
                    if rule:
                        commission = so.currency_id.round(line.price_subtotal * rule.rate / 100.0)
                        comm_by_rule[rule] += commission

                # cap by rule
                for r, amount in comm_by_rule.items():
                    if r.is_capped:
                        amount = min(amount, r.max_commission)
                        comm_by_rule[r] = amount

                so.commission = sum(comm_by_rule.values())

    @api.depends('partner_id', 'referrer_id', 'referrer_id.commission_plan_id')
    def _compute_commission_plan(self):
        for so in self:
            if so.state in ('draft', 'sent'):
                so.commission_plan_id = so.referrer_id.commission_plan_id

    def _set_commission_plan(self):
        for so in self:
            # forward commission plan to subscription
            subs = so.mapped('order_line.subscription_id')
            subs.filtered(lambda s: s.stage_category == 'progress' and not s.commission_plan_frozen).sudo().write({
                'commission_plan_id': so.commission_plan_id.id,
            })

    def _prepare_invoice(self):
        invoice_vals = super(SaleOrder, self)._prepare_invoice()
        if self.referrer_id:
            invoice_vals.update({
                'referrer_id': self.referrer_id.id,
            })
        return invoice_vals

    def _prepare_subscription_data(self, template):
        values = super(SaleOrder, self)._prepare_subscription_data(template)
        if self.referrer_id:
            values.update({
                'referrer_id': self.referrer_id.id,
                'commission_plan_id': self.commission_plan_id.id,
                'commission_plan_frozen': self.referrer_id.commission_plan_id != self.commission_plan_id,
            })
        return values


class SaleAdvancePaymentInv(models.TransientModel):
    _inherit = 'sale.advance.payment.inv'

    def _prepare_invoice_values(self, order, name, amount, so_line):
        res = super()._prepare_invoice_values(order, name, amount, so_line)
        res['referrer_id'] = order.referrer_id.id

        return res
